// Description: 
// By Sarah Dean
// Email: sdean12@sdean12.org
// WWW:   http://www.FreeOTFE.org/
//
// -----------------------------------------------------------------------------
//


#include "FreeOTFECypherImpl.h"
#include "FreeOTFECypherTwofish_HifnCS.h"

#include "FreeOTFEDebug.h"
#include "FreeOTFElib.h"


// Include Hi/fn and Counterpane Systems Twofish library...
#include "AES.H"


// =========================================================================
// Cypher driver init function
// devExt - The device extension to be initialized
NTSTATUS
ImpCypherDriverExtDetailsInit(
    IN OUT DEVICE_EXTENSION* devExt
)
{
    NTSTATUS status = STATUS_SUCCESS;
    int idx = -1;

    DEBUGOUTCYPHERIMPL(DEBUGLEV_ENTER, ("ImpCypherDriverExtDetailsInit\n"));


    // -- POPULATE DRIVER IDENTIFICATION --
    RtlZeroMemory(devExt->DriverTitle, sizeof(devExt->DriverTitle));
    RtlCopyMemory(
                  devExt->DriverTitle,
                  DRIVER_TITLE,
                  strlen(DRIVER_TITLE)
                 );

    devExt->DriverGUID = DRIVER_GUID;
    devExt->DriverVersionID = DRIVER_CYPHER_VERSION;


    // -- POPULATE cyphers SUPPORTED --
    devExt->CypherDetails = ExAllocatePool(
                                         NonPagedPool,
                                         (sizeof(CYPHER) * CYPHERS_SUPPORTED)
                                        );    

    devExt->CypherCount = CYPHERS_SUPPORTED;


    // -- Twofish-128 --
    idx++;
    RtlZeroMemory(devExt->CypherDetails[idx].Title, MAX_CYPHER_TITLE);
    RtlCopyMemory(
                  devExt->CypherDetails[idx].Title,
                  DRIVER_CIPHER_TITLE_TWOFISH_128,
                  strlen(DRIVER_CIPHER_TITLE_TWOFISH_128)
                 );

    devExt->CypherDetails[idx].Mode       = CYPHER_MODE_CBC;
    devExt->CypherDetails[idx].KeySize    = 128;
    devExt->CypherDetails[idx].BlockSize  = BLOCK_SIZE;  // From TWOFISH.C
    devExt->CypherDetails[idx].VersionID  = DRIVER_CYPHER_IMPL_VERSION;
    devExt->CypherDetails[idx].CypherGUID = CIPHER_GUID_TWOFISH_128;


    // -- Twofish-192 --
    idx++;
    RtlZeroMemory(devExt->CypherDetails[idx].Title, MAX_CYPHER_TITLE);
    RtlCopyMemory(
                  devExt->CypherDetails[idx].Title,
                  DRIVER_CIPHER_TITLE_TWOFISH_192,
                  strlen(DRIVER_CIPHER_TITLE_TWOFISH_192)
                 );

    devExt->CypherDetails[idx].Mode       = CYPHER_MODE_CBC;
    devExt->CypherDetails[idx].KeySize    = 192;
    devExt->CypherDetails[idx].BlockSize  = BLOCK_SIZE;  // From TWOFISH.C
    devExt->CypherDetails[idx].VersionID  = DRIVER_CYPHER_IMPL_VERSION;
    devExt->CypherDetails[idx].CypherGUID = CIPHER_GUID_TWOFISH_192;


    // -- Twofish-256 --
    idx++;
    RtlZeroMemory(devExt->CypherDetails[idx].Title, MAX_CYPHER_TITLE);
    RtlCopyMemory(
                  devExt->CypherDetails[idx].Title,
                  DRIVER_CIPHER_TITLE_TWOFISH_256,
                  strlen(DRIVER_CIPHER_TITLE_TWOFISH_256)
                 );

    devExt->CypherDetails[idx].Mode       = CYPHER_MODE_CBC;
    devExt->CypherDetails[idx].KeySize    = 256;
    devExt->CypherDetails[idx].BlockSize  = BLOCK_SIZE;  // From TWOFISH.C
    devExt->CypherDetails[idx].VersionID  = DRIVER_CYPHER_IMPL_VERSION;
    devExt->CypherDetails[idx].CypherGUID = CIPHER_GUID_TWOFISH_256;


    DEBUGOUTCYPHERIMPL(DEBUGLEV_EXIT, ("ImpCypherDriverExtDetailsInit\n"));

    return status;
}


// =========================================================================
// Cypher driver cleardown function
// devExt - The device extension to be cleared down
NTSTATUS
ImpCypherDriverExtDetailsCleardown(    
    IN OUT DEVICE_EXTENSION* devExt
)
{
    NTSTATUS status = STATUS_SUCCESS;

    DEBUGOUTCYPHERIMPL(DEBUGLEV_ENTER, ("ImpCypherDriverExtDetailsCleardown\n"));

    if (devExt->CypherDetails != NULL)
        {
        ExFreePool(devExt->CypherDetails);
        }

    devExt->CypherDetails = NULL;
    devExt->CypherCount = 0;

    DEBUGOUTCYPHERIMPL(DEBUGLEV_EXIT, ("ImpCypherDriverExtDetailsCleardown\n"));

    return status;
}


// =========================================================================
// Encryption function
// Note: CyphertextLength must be set to the size of the CyphertextData buffer on
//       entry; on exit, this will be set to the size of the buffer used.
NTSTATUS
ImpCypherEncryptData(
    IN      GUID* CypherGUID,
    IN      int KeyLength,  // In bits
    IN      char *Key,
    IN      char *KeyASCII,  // ASCII representation of "Key"
    IN      int IVLength,  // In bits
    IN      char *IV,
    IN      int PlaintextLength,  // In bytes
    IN      char *PlaintextData,
    OUT     char *CyphertextData
)
{
    NTSTATUS status = STATUS_SUCCESS;
    cipherInstance cipher;
    // NOTE: We can't do "keyInstance key;" as this causes a __chkstk error on
    //       building. For that reason, we have to manually allocate the memory
    //       outselves
    keyInstance* key;
    int countBitsProcessed;
    int i;
    char* asciizIV;
    unsigned int asciizIVLen;

    DEBUGOUTCYPHERIMPL(DEBUGLEV_ENTER, ("ImpCypherEncryptData\n"));

    if (!(
		  (IsEqualGUID(&CIPHER_GUID_TWOFISH_128, CypherGUID)) || 
          (IsEqualGUID(&CIPHER_GUID_TWOFISH_192, CypherGUID)) ||
          (IsEqualGUID(&CIPHER_GUID_TWOFISH_256, CypherGUID))
		 ))
        {
		DEBUGOUTCYPHERIMPL(DEBUGLEV_ERROR, ("Unsupported cipher GUID passed in.\n"));
        status = STATUS_INVALID_PARAMETER;
        }

    key = ExAllocatePool(
                         NonPagedPool,
                         sizeof(keyInstance)
                        );    
    RtlZeroMemory(key, sizeof(keyInstance));

    RtlZeroMemory(&cipher, sizeof(cipherInstance));

    if NT_SUCCESS(status)
        {
        if ( (IV == NULL) && (IVLength != 0) )
            {
            // Sanity check failed...
            DEBUGOUTCYPHERIMPL(DEBUGLEV_INFO, ("Inconsistant: NULL passed as %d bit long IV\n", IVLength));
            status = STATUS_INVALID_PARAMETER;
            }
        else if ( (IV == NULL) || (IVLength == 0) )
            {
            // Note: The IV is NULL; it's zero's by the RtlZeroMemory above
            DEBUGOUTCYPHERIMPL(DEBUGLEV_INFO, ("Using NULL IV\n"));
            asciizIV = NULL;
            }
        else
            {
            DEBUGOUTCYPHERIMPL(DEBUGLEV_INFO, ("Using specific IV\n"));

            // Convert the IV to it's ASCIIZ representation
            // Horrific, but that's the cypher AES API for you...
            asciizIVLen = (IVLength / 4) + 1; // Convert bits into nibbles,
                                              // and add a NULL terminator

            asciizIV = ExAllocatePool(NonPagedPool, asciizIVLen);
            status = ConvertDataToASCIIRep(IVLength, IV, asciizIV);
            if (!(NT_SUCCESS(status)))
                {
                DEBUGOUTCYPHERIMPL(DEBUGLEV_ERROR, ("Unable to convert IV to it's ASCIIZ representation\n"));
                }
            }

        }

    if NT_SUCCESS(status)
        {
        if (cipherInit(&cipher, MODE_CBC, asciizIV) != TRUE)
            {
            DEBUGOUTCYPHERIMPL(DEBUGLEV_ERROR, ("Could not cipherInit.\n"));
            status = STATUS_INTERNAL_ERROR;
            }
        }


    if NT_SUCCESS(status)
        {
        if (makeKey(key, DIR_DECRYPT, KeyLength, KeyASCII) != TRUE)
            {
            DEBUGOUTCYPHERIMPL(DEBUGLEV_ERROR, ("Could not makeKey.\n"));
            status = STATUS_INTERNAL_ERROR;
            }
        }

    if NT_SUCCESS(status)
        {  
        countBitsProcessed = blockEncrypt(&cipher, key, (BYTE*)PlaintextData, (PlaintextLength * 8), (BYTE*)CyphertextData);
        if (countBitsProcessed != (PlaintextLength * 8))
            {
            DEBUGOUTCYPHERIMPL(DEBUGLEV_ERROR, ("Could not encrypt/decrypt all; expected: %d; got: %d\n", (PlaintextLength * 8), countBitsProcessed));
            status = STATUS_INTERNAL_ERROR;
            }
        }


    // Nuke the key and cipher instance...
    SecZeroMemory(&cipher, sizeof(cipherInstance));

    if (asciizIV != NULL)
        {
        SecZeroMemory(asciizIV, sizeof(asciizIVLen));
        ExFreePool(asciizIV);
        }

    SecZeroMemory(key, sizeof(keyInstance));
    ExFreePool(key);

    DEBUGOUTCYPHERIMPL(DEBUGLEV_EXIT, ("ImpCypherEncryptData\n"));

    return status;
}


// =========================================================================
// Decryption function
// Note: PlaintextLength must be set to the size of the PlaintextData buffer on
//       entry; on exit, this will be set to the size of the buffer used.
NTSTATUS
ImpCypherDecryptData(
    IN      GUID* CypherGUID,
    IN      int KeyLength,  // In bits
    IN      char *Key,
    IN      char *KeyASCII,  // ASCII representation of "Key"
    IN      int IVLength,  // In bits
    IN      char *IV,
    IN      int CyphertextLength,  // In bytes
    IN      char *CyphertextData,
    OUT     char *PlaintextData
)
{
    NTSTATUS status = STATUS_SUCCESS;
    cipherInstance cipher;
    // NOTE: We can't do "keyInstance key;" as this causes a __chkstk error on
    //       building. For that reason, we have to manually allocate the memory
    //       outselves
    keyInstance* key;
    int countBitsProcessed;
    int i;
    char* asciizIV;
    unsigned int asciizIVLen;

    DEBUGOUTCYPHERIMPL(DEBUGLEV_ENTER, ("ImpCypherDecryptData\n"));

    if (!(
		  (IsEqualGUID(&CIPHER_GUID_TWOFISH_128, CypherGUID)) || 
          (IsEqualGUID(&CIPHER_GUID_TWOFISH_192, CypherGUID)) ||
          (IsEqualGUID(&CIPHER_GUID_TWOFISH_256, CypherGUID))
		 ))
		{
		DEBUGOUTCYPHERIMPL(DEBUGLEV_ERROR, ("Unsupported cipher GUID passed in.\n"));
        status = STATUS_INVALID_PARAMETER;
        }

    key = ExAllocatePool(
                         NonPagedPool,
                         sizeof(keyInstance)
                        );    
    RtlZeroMemory(key, sizeof(keyInstance));

    RtlZeroMemory(&cipher, sizeof(cipherInstance));

    if NT_SUCCESS(status)
        {
        if ( (IV == NULL) && (IVLength != 0) )
            {
            // Sanity check failed...
            DEBUGOUTCYPHERIMPL(DEBUGLEV_INFO, ("Inconsistant: NULL passed as %d bit long IV\n", IVLength));
            status = STATUS_INVALID_PARAMETER;
            }
        else if ( (IV == NULL) || (IVLength == 0) )
            {
            // Note: The IV is NULL; it's zero's by the RtlZeroMemory above
            DEBUGOUTCYPHERIMPL(DEBUGLEV_INFO, ("Using NULL IV\n"));
            asciizIV = NULL;
            }
        else
            {
            DEBUGOUTCYPHERIMPL(DEBUGLEV_INFO, ("Using specific IV\n"));

            // Convert the IV to it's ASCIIZ representation
            // Horrific, but that's the cypher AES API for you...
            asciizIVLen = (IVLength / 4) + 1; // Convert bits into nibbles,
                                              // and add a NULL terminator

            asciizIV = ExAllocatePool(NonPagedPool, asciizIVLen);
            status = ConvertDataToASCIIRep(IVLength, IV, asciizIV);
            if (!(NT_SUCCESS(status)))
                {
                DEBUGOUTCYPHERIMPL(DEBUGLEV_ERROR, ("Unable to convert IV to it's ASCIIZ representation\n"));
                }
            }

        }

    if NT_SUCCESS(status)
        {
        if (cipherInit(&cipher, MODE_CBC, asciizIV) != TRUE)
            {
            DEBUGOUTCYPHERIMPL(DEBUGLEV_ERROR, ("Could not cipherInit.\n"));
            status = STATUS_INTERNAL_ERROR;
            }
        }


    if NT_SUCCESS(status)
        {
        if (makeKey(key, DIR_DECRYPT, KeyLength, KeyASCII) != TRUE)
            {
            DEBUGOUTCYPHERIMPL(DEBUGLEV_ERROR, ("Could not makeKey.\n"));
            status = STATUS_INTERNAL_ERROR;
            }
        }

    if NT_SUCCESS(status)
        {  
        countBitsProcessed = blockDecrypt(&cipher, key, (BYTE*)CyphertextData, (CyphertextLength * 8), (BYTE*)PlaintextData);
        if (countBitsProcessed != (CyphertextLength * 8))
            {
            DEBUGOUTCYPHERIMPL(DEBUGLEV_ERROR, ("Could not encrypt/decrypt all; expected: %d; got: %d\n", (CyphertextLength * 8), countBitsProcessed));
            status = STATUS_INTERNAL_ERROR;
            }
        }


    // Nuke the key and cipher instance...
    SecZeroMemory(&cipher, sizeof(cipherInstance));

    if (asciizIV != NULL)
        {
        SecZeroMemory(asciizIV, sizeof(asciizIVLen));
        ExFreePool(asciizIV);
        }

    SecZeroMemory(key, sizeof(keyInstance));
    ExFreePool(key);

    DEBUGOUTCYPHERIMPL(DEBUGLEV_EXIT, ("ImpCypherDecryptData\n"));

    return status;
}


// =========================================================================
// =========================================================================

