// Description: FreeOTFE Cypher Device Driver API
// By Sarah Dean
// Email: sdean12@sdean12.org
// WWW:   http://www.FreeOTFE.org/
//
// -----------------------------------------------------------------------------
//


#ifndef _FreeOTFECypherAPI_H
#define _FreeOTFECypherAPI_H   1

// FreeOTFEAPI.h is included for the definition of DEVICE_FREEOTFE_ROOT, etc
#include "FreeOTFEAPI.h"


// Required for definition of ULONG
#include <windef.h>




// Following is required for the "_T" macro
/// #include <tchar.h>

#include <initguid.h>  // Required for DEFINE_GUID and other GUID defintions

// #define _T(x) L ## x

// =========================================================================
// Const definitions

#define MAX_CYPHER_TITLE 256

// Maximum cypher blocksize
// This should be set to the MAXIMUM cypher blocksize possible.
// This is used for allocating storage for IVs; i.e.:
//   1) In the FreeOTFE driver when generating sector IVs
//   2) So that the length of the DIOC passed to the cypher driver is more deterministic, which
//      makes creating and populating this struct easier
// If a cypher uses a blocksize greater than this value (e.g. a cypher driver which 
// uses an arbitary blocksize), then attempts to encrypt/decrypt may fail if the IVs passed to
// the encryption/decryption routines may not be large enough.
// Note: This value is in *bits*
#define FREEOTFE_MAX_CYPHER_BLOCKSIZE    512



#define DEVICE_CYPHER_BASE_NAME          L"\\FreeOTFECypher"

#define DEVICE_CYPHER_DIR_NAME           DEVICE_FREEOTFE_ROOT L"\\Cypher"
#define DEVICE_CYPHER_SYMLINK_DIR_NAME   DEVICE_SYMLINK_FREEOTFE_ROOT
#define DEVICE_CYPHER_SYMLINK_PREFIX     DEVICE_CYPHER_SYMLINK_DIR_NAME DEVICE_CYPHER_BASE_NAME



#define IOCTL_FreeOTFECypher_IDENTIFYDRIVER    CTL_CODE(             \
                                                FILE_DEVICE_UNKNOWN, \
                                                0x800,               \
                                                METHOD_BUFFERED,     \
                                                FILE_READ_DATA       \
                                               )

#define IOCTL_FreeOTFECypher_IDENTIFYSUPPORTED    CTL_CODE(          \
                                                FILE_DEVICE_UNKNOWN, \
                                                0x801,               \
                                                METHOD_BUFFERED,     \
                                                FILE_READ_DATA       \
                                               )

#define IOCTL_FreeOTFECypher_INTLDETAILS  CTL_CODE(                  \
                                                FILE_DEVICE_UNKNOWN, \
                                                0x802,               \
                                                METHOD_BUFFERED,     \
                                                FILE_READ_DATA       \
                                               )

#define IOCTL_FreeOTFECypher_ENCRYPT      CTL_CODE(                  \
                                                FILE_DEVICE_UNKNOWN, \
                                                0x803,               \
                                                METHOD_BUFFERED,     \
                                                FILE_READ_DATA       \
                                               )

#define IOCTL_FreeOTFECypher_DECRYPT      CTL_CODE(                  \
                                                FILE_DEVICE_UNKNOWN, \
                                                0x804,               \
                                                METHOD_BUFFERED,     \
                                                FILE_READ_DATA       \
                                               )



// =========================================================================
// Type definitions

typedef enum _CYPHER_MODE {
    CYPHER_MODE_NONE    =    0,
    CYPHER_MODE_ECB     =    1,
    CYPHER_MODE_CBC     =    2,
    CYPHER_MODE_UNKNOWN = 9999
} CYPHER_MODE, *PCYPHER_MODE;

typedef struct _CYPHER {
    GUID CypherGUID;
    char Title[MAX_CYPHER_TITLE];
    CYPHER_MODE Mode;
    // KeySize = -1 for *any* KeySize
    int KeySize;  // In bits
    // BlockSize = -1 for *any* BlockSize
    int BlockSize;  // In bits
    ULONG VersionID;
} CYPHER, *PCYPHER;


typedef struct _DIOC_CYPHER_IDENTIFYDRIVER {    
    GUID DriverGUID;
    char Title[MAX_CYPHER_TITLE];
    ULONG VersionID;
    unsigned int CountCyphers;
} DIOC_CYPHER_IDENTIFYDRIVER, *PDIOC_CYPHER_IDENTIFYDRIVER;


typedef struct _DIOC_CYPHER_IDENTIFYSUPPORTED {    
    unsigned int BufCount;  // The number of *elements* in the "Cyphers" array
    CYPHER Cyphers[1];  // Variable length
} DIOC_CYPHER_IDENTIFYSUPPORTED, *PDIOC_CYPHER_IDENTIFYSUPPORTED;


typedef struct _DIOC_CYPHER_DATA_IN {
    GUID CypherGUID;
    int KeyLength;  // In *bits*
    int IVLength;  // In *bits*
                   // Note: Not all cyphers use an IV (e.g. in EBC mode)
                   //       If specified though, IVLength, and IV, should 
                   //       be equal to the blocksize of the cypher being used
    int DataLength;  // In *bytes*
    char Key[1];  // Variable length
    char IV[1];  // Variable length
    char Data[1];  // Variable length
} DIOC_CYPHER_DATA_IN, *PDIOC_CYPHER_DATA_IN;


// When used, this buffer must be the same size as the "Data" member
// of DIOC_CYPHER_DATA_IN
typedef struct _DIOC_CYPHER_DATA_OUT {
    char Data[1];  // Variable length
} DIOC_CYPHER_DATA_OUT, *PDIOC_CYPHER_DATA_OUT;



// This definition must be kept in sync with "ImpCypherEncryptData"
typedef NTSTATUS (* PDataEncryptFn)(
                                    GUID*,
                                    int, 
                                    char*, 
                                    char*, 
                                    int, 
                                    char*, 
                                    int, 
                                    char*, 
                                    char*
                                   );


// This definition must be kept in sync with "ImpCypherDecryptData"
typedef NTSTATUS (* PDataDecryptFn)(
                                    GUID*,
                                    int, 
                                    char*, 
                                    char*, 
                                    int, 
                                    char*, 
                                    int, 
                                    char*, 
                                    char*
                                   );

// This definition must be kept in sync with "GetCypherDetails"
typedef NTSTATUS (* PCypherDetailsFn)(
                                    GUID*,
                                    CYPHER*
                                   );

// Internal details
typedef struct _DIOC_CYPHER_INTL_DETAILS {
    PCypherDetailsFn FnCypherDetails;
    PDataEncryptFn   FnEncrypt;
    PDataDecryptFn   FnDecrypt;
} DIOC_CYPHER_INTL_DETAILS, *PDIOC_CYPHER_INTL_DETAILS;




// =========================================================================
// =========================================================================

#endif

